/*!
 *  Copyright 2016 Ron Buckton (rbuckton@chronicles.org)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const fs = require("fs");
const path = require("path");
const gutil = require("gulp-util");
const ecmarkup_1 = require("ecmarkup");
const stream_1 = require("stream");
const prex_1 = require("prex");
const Vinyl = require("vinyl");
function ecmarkup(opts) {
    return new ecmarkup.EcmarkupTransform(opts);
}
(function (ecmarkup) {
    class EcmarkupTransform extends stream_1.Transform {
        constructor(opts = {}) {
            super({ objectMode: true });
            this._queue = new prex_1.AsyncQueue();
            this._countdown = new prex_1.CountdownEvent(1);
            this._cache = new Map();
            this._opts = {
                js: pluck(opts, "js"),
                css: pluck(opts, "css"),
                biblio: pluck(opts, "biblio")
            };
            this._emuOpts = opts;
            this._waitForWrite();
        }
        _write(file, enc, cb) {
            if (file.isNull()) {
                return cb();
            }
            if (file.isStream()) {
                throw new gutil.PluginError("gulp-emu", "Stream not supported.");
            }
            // cache the file contents
            this._cache.set(file.path, file.contents.toString("utf8"));
            // put an entry into the queue for the transformation
            this._enqueue(this._buildAsync(file));
            cb();
        }
        _flush(cb) {
            const files = [];
            const css = this._opts.css;
            if (css) {
                const dest = typeof css === "string" ? css : "elements.css";
                files.push({ src: [require.resolve("ecmarkup/css/elements.css")], dest });
            }
            const js = this._opts.js;
            if (js) {
                if (typeof js === "string") {
                    files.push({
                        src: [
                            require.resolve("ecmarkup/js/menu.js"),
                            require.resolve("ecmarkup/js/findLocalReferences.js")
                        ],
                        dest: js
                    });
                }
                else {
                    files.push({ src: [require.resolve("ecmarkup/js/menu.js")], dest: "menu.js" });
                    files.push({ src: [require.resolve("ecmarkup/js/findLocalReferences.js")], dest: "findLocalReferences.js" });
                }
            }
            if (files.length) {
                this._enqueue(this._readFilesAsync(files));
            }
            this._countdown.signal();
            this._countdown.wait()
                .then(() => cb())
                .catch(e => this.emit("error", e));
        }
        _enqueue(promise) {
            this._countdown.add();
            this._queue.put(promise);
        }
        _finishWrite(files) {
            for (const file of files) {
                this.push(file);
            }
            this._countdown.signal();
            this._waitForWrite();
        }
        _waitForWrite() {
            this._queue.get().then(file => this._finishWrite(file), e => this.emit("error", e));
        }
        _buildAsync(file) {
            return __awaiter(this, void 0, void 0, function* () {
                const files = [];
                const spec = yield ecmarkup_1.build(file.path, path => this._readFileAsync(path), this._emuOpts);
                if (spec) {
                    if (this._opts.biblio) {
                        const dirname = path.dirname(file.path);
                        const extname = path.extname(file.path);
                        const basename = path.basename(file.path, extname);
                        const biblio = new Vinyl({
                            path: path.join(dirname, basename + ".biblio.json"),
                            base: file.base,
                            contents: new Buffer(JSON.stringify(spec.exportBiblio()), "utf8")
                        });
                        files.push(biblio);
                    }
                    const html = spec.toHTML();
                    file.contents = new Buffer(html, "utf8");
                    files.push(file);
                }
                return files;
            });
        }
        _readFilesAsync(files) {
            return __awaiter(this, void 0, void 0, function* () {
                return yield Promise.all(files.map(file => this._mergeFilesAsync(file)));
            });
        }
        _mergeFilesAsync(file) {
            return __awaiter(this, void 0, void 0, function* () {
                const srcContents = yield Promise.all(file.src.map(src => this._readFileAsync(src)));
                const contents = srcContents.join("");
                const base = path.dirname(file.src[0]);
                return new Vinyl({
                    path: path.join(base, file.dest),
                    base,
                    contents: new Buffer(contents, "utf8")
                });
            });
        }
        _readFileAsync(path) {
            return __awaiter(this, void 0, void 0, function* () {
                let contents = this._cache.get(path);
                if (!contents) {
                    contents = yield readFile(path, "utf8");
                    this._cache.set(path, contents);
                }
                return contents;
            });
        }
    }
    ecmarkup.EcmarkupTransform = EcmarkupTransform;
})(ecmarkup || (ecmarkup = {}));
function readFile(file, encoding) {
    return new Promise((resolve, reject) => {
        fs.readFile(file, encoding, (err, data) => err ? reject(err) : resolve(data));
    });
}
function pluck(obj, key) {
    const value = obj[key];
    delete obj[key];
    return value;
}
module.exports = ecmarkup;
